/**
 * Created by AlexanderC on 5/25/15.
 */
'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Runtime = undefined;

var _Interface = require('../../OOP/Interface');

var _Response = require('./Response');

var _ErrorResponse = require('./ErrorResponse');

var _Request = require('./Request');

var _InvalidCognitoIdentityException = require('./Exception/InvalidCognitoIdentityException');

var _MissingUserContextException = require('./Exception/MissingUserContextException');

var _Context = require('./Context');

var _Sandbox = require('../../Runtime/Sandbox');

var _Resolver = require('./Resolver');

var _awsSdk = require('aws-sdk');

var _awsSdk2 = _interopRequireDefault(_awsSdk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Lambda runtime context
 */
class Runtime extends _Interface.Interface {
  /**
   * @param {Object} kernel
   */
  constructor(kernel) {
    super(['handle']);

    this._kernel = kernel;
    this._request = null;
    this._context = null;
    this._resolver = null;

    this._loggedUserId = null;
    this._forceUserIdentity = false;
    this._contextSent = false;

    this._calleeConfig = null;

    this._fillDenyMissingUserContextOption();
  }

  /**
   * @returns {null|Context}
   */
  get context() {
    return this._context;
  }

  /**
   * @returns {null|Resolver}
   */
  get resolver() {
    return this._resolver;
  }

  /**
   * @returns {Boolean}
   */
  get contextSent() {
    return this._contextSent;
  }

  /**
   * @returns {String}
   */
  get loggedUserId() {
    return this._loggedUserId;
  }

  /**
   * @returns {Boolean}
   */
  get forceUserIdentity() {
    return this._forceUserIdentity;
  }

  /**
   * @returns {Object}
   */
  get kernel() {
    return this._kernel;
  }

  /**
   * @returns {Function}
   */
  get lambda() {
    let _this = this;

    return function (event, context) {
      _this.run(event, context);
    };
  }

  /**
   * @param {String} schemaName
   * @param {Function} cb
   * @returns {Runtime}
   */
  validateInput(schemaName, cb) {
    let validation = this._kernel.get('validation');

    validation.validateRuntimeInput(this, schemaName, cb);

    return this;
  }

  /**
   * @param {Object} event
   * @param {Object} context
   * @param {Function} lambdaCallback
   * @returns {Runtime}
   */
  run(event, context, lambdaCallback = null) {
    this._context = new _Context.Context(context);
    this._request = new _Request.Request(event);

    if (lambdaCallback) {
      this._resolver = new _Resolver.Resolver(lambdaCallback);

      this._resolver.registerSucceedCallback(this.kernel.kernelCleanup.bind(this.kernel));
    }

    this._context.waitForEmptyEventLoop();

    this.logService.rumLog({
      service: 'deep-core',
      resourceType: 'Lambda',
      resourceId: this._context.invokedFunctionArn,
      eventName: 'Run'
    });

    new _Sandbox.Sandbox(() => {
      this._fillUserContext().then(() => {
        if (!this._loggedUserId && this._forceUserIdentity) {
          throw new _MissingUserContextException.MissingUserContextException();
        }

        this._initDBPartitionKey();

        this._runValidate(this.validationSchema, validatedData => {
          if (this._context._lambdaContext.overheadSubSegment) {
            this._context._lambdaContext.overheadSubSegment.close();
          }

          this.handle(validatedData);
        });
      }).catch(e => this.createError(e).send());
    }).fail(error => {
      this.createError(error).send();
    }).run();

    return this;
  }

  /**
   * @param {String} validationSchema
   * @param {Function} cb
   * @private
   */
  _runValidate(validationSchema, cb) {
    if (!validationSchema) {
      cb(this._request);
      return;
    }

    let validationSchemaName = validationSchema;

    if (typeof validationSchema !== 'string') {
      validationSchemaName = this._injectValidationSchema(validationSchema);
    }

    this.validateInput(validationSchemaName, cb);
  }

  /**
   * @param {Object} schema
   * @param {String|null} name
   * @returns {String}
   * @private
   */
  _injectValidationSchema(schema, name = null) {
    let validation = this._kernel.get('validation');

    name = name || `DeepHandlerValidation_${new Date().getTime()}`;

    validation.setGuessSchema(name, schema);

    return name;
  }

  /**
   * @param {String|Error|*} error
   * @returns {ErrorResponse}
   */
  createError(error) {
    return new _ErrorResponse.ErrorResponse(this, error);
  }

  /**
   * @param {Object} data
   * @returns {Response}
   */
  createResponse(data) {
    return new _Response.Response(this, data);
  }

  /**
   * @param {Object} data
   * @param {Number} ttl
   * @param {Function} callback
   */
  createCachedResponse(data, ttl = 0, callback = response => response.send()) {
    let publicCache = this._kernel.get('cache').shared;
    let publicCacheKey = publicCache.buildKeyFromLambdaRuntime(this);
    let response = new _Response.Response(this, data);

    publicCache.assure(publicCacheKey, { body: data }, ttl, () => {
      callback(response);
    });
  }

  /**
   * @returns {String}
   */
  get calleeName() {
    if (this._context && this._context.has('invokedFunctionArn')) {
      let calleeArn = this._context.getOption('invokedFunctionArn');

      return calleeArn.replace(/^.+:function:/i, '');
    }

    return null;
  }

  /**
   * @returns {null|Object}
   */
  get calleeConfig() {
    if (!this._calleeConfig && this._context && this._kernel && this._context.has('invokedFunctionArn')) {

      let resource = this._kernel.get('resource');
      let calleeArn = this._context.getOption('invokedFunctionArn');

      this._calleeConfig = resource.getActionConfig(calleeArn);
    }

    return this._calleeConfig;
  }

  /**
   * @returns {String}
   */
  get validationSchema() {
    return this.calleeConfig ? this.calleeConfig.validationSchema || null : null;
  }

  /**
   * @returns {Request}
   */
  get request() {
    return this._request;
  }

  /**
   * @returns {Object}
   */
  get securityService() {
    return this.kernel.get('security');
  }

  /**
   * @returns {DB}
   */
  get dbService() {
    return this.kernel.get('db');
  }

  /**
   * @returns {Object}
   */
  get logService() {
    return this.kernel.get('log');
  }

  /**
   * @private
   */
  _fillDenyMissingUserContextOption() {
    if (this._kernel.config.hasOwnProperty('forceUserIdentity')) {
      this._forceUserIdentity = this._kernel.config.forceUserIdentity;
    }
  }

  /**
   * @returns {Runtime}
   * @private
   */
  _initDBPartitionKey() {
    if (!this.kernel.accountMicroservice || !this.forceUserIdentity) {
      return this;
    }

    let db = this.dbService;
    let user = this.securityService.token && this.securityService.token.user;
    let partitionKey = Runtime.DB_ANONYMOUS_PARTITION;

    if (user && user.ActiveAccount) {
      partitionKey = user.ActiveAccount.Id;
    }

    db.setDynamoDBPartitionKey(partitionKey);

    return this;
  }

  /**
   * Retrieves logged user id from lambda context
   *
   * @private
   * @returns {Promise}
   */
  _fillUserContext() {
    if (this._context && this._context.identity && this._context.identity.cognitoIdentityPoolId && this._context.identity.cognitoIdentityId) {
      let identityPoolId = this._context.identity.cognitoIdentityPoolId;

      if (this.securityService.identityPoolId !== identityPoolId && !this._kernel.container.localBackend) {

        return Promise.reject(new _InvalidCognitoIdentityException.InvalidCognitoIdentityException(identityPoolId));
      }

      // inject lambda context into security service
      // and instantiate security token without loading user credentials
      return this.securityService.warmupBackendLogin(this._context).then(credentials => {
        let dbService = this.dbService;

        _awsSdk2.default.config.credentials = credentials;
        credentials && dbService.overwriteCredentials(credentials);

        this._loggedUserId = this._context.identity.cognitoIdentityId;
      });
    }

    return Promise.resolve(null);
  }

  /**
   * @returns {String}
   */
  static get DB_ANONYMOUS_PARTITION() {
    return 'anonymous';
  }

  /**
   * @returns {String}
   */
  static get VALIDATION_SCHEMAS_DIR() {
    return '__deep_validation_schemas__';
  }
}
exports.Runtime = Runtime;